from http import HTTPStatus
from unittest.mock import AsyncMock, patch

import parameterized
from synapse.api.constants import CREATOR_POWER_LEVEL, EventTypes
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.rest.types import CreateManagedRoomRequest
from tests.utils.module_api_testcase import ModuleApiTestCase


@parameterized.parameterized_class(("room_version",), [("10",), ("12",)])
class TestPowerLevelHook(ModuleApiTestCase):
    """
    Test that trying to change the invariant power levels of an existing room is:
    1. Prohibited by a non-room creator admin
    2. Isn't allowed if any other user is given the same power level of the room creator
    3. Isn't allowed if any sensitive event types are changed
    4. Isn't allowed if any membership actions are changed
    5. Is allowed for normal room administrator behavior (nothing unexpected broke)
    """

    room_version: str
    CREATE_PATH = "/_famedlyControl/v1/managedRooms/createRoom"

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.non_admin = self.register_user("non_admin", "password", admin=False)
        self.non_admin_token = self.login("non_admin", "password")

    def _create_managed_room(
        self, name: str = "PL Test Room", groups: list[str] | None = None
    ) -> str:
        config = CreateManagedRoomRequest(
            room_alias_name="pl_test_room",
            name=name,
            room_version=self.room_version,
            groups=groups or ["test_group"],
        )
        with (
            patch(
                "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
                new_callable=AsyncMock,
                return_value=[],
            ),
        ):
            channel = self.make_request(
                method="POST",
                path=self.CREATE_PATH,
                content=config.model_dump(),
                access_token=self.creator_access_token,
                shorthand=False,
            )
        assert channel.code == HTTPStatus.OK, channel.result
        return channel.json_body["room_id"]

    def _get_power_levels(self, room_id: str) -> dict:
        return self.helper.get_state(
            room_id, EventTypes.PowerLevels, self.creator_access_token
        )

    def _set_power_levels(
        self, room_id: str, content: dict, expect_code: int = HTTPStatus.OK
    ) -> dict:
        return self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            content,
            self.creator_access_token,
            expect_code=expect_code,
        )

    def test_reject_other_user_equal_to_admin(self) -> None:
        """Setting another user's PL equal to the admin's should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = CREATOR_POWER_LEVEL - 1
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_other_user_higher_than_admin(self) -> None:
        """Setting another user's PL higher than the admin's should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = CREATOR_POWER_LEVEL
        # CREATOR_POWER_LEVEL is non-canonicaljson, so is rejected
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.BAD_REQUEST)

    def test_reject_room_creator_cannot_lower_own_power_level(self) -> None:
        """Setting a room creator's power level too low should be rejected"""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.creator] = CREATOR_POWER_LEVEL - 2

        # Room versions above "11" do not allow for the room creator to exist in the
        # power level event's 'users' object, as they will always have an 'infinite'
        # level. The response there is different for rooms "11" and below. However, the
        # validation done on the power level event takes place before that check, so
        # instead of raising a 400 it will be the 403
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    @parameterized.parameterized.expand([("ban",), ("kick",), ("invite",)])
    def test_membership_action_power_level_equal_to_admin_no_op(
        self, action_level: str
    ) -> None:
        """Setting specific action PL at admin user's PL should be allowed."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl[action_level] = CREATOR_POWER_LEVEL - 1
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

    @parameterized.parameterized.expand([("ban",), ("kick",), ("invite",)])
    def test_membership_action_power_level_below_admin_rejected(
        self, action_level: str
    ) -> None:
        """Setting specific action PL below admin user's PL should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl[action_level] = CREATOR_POWER_LEVEL - 2
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_admin_can_update_users_power_levels(self) -> None:
        """The room admin should be able to update power levels."""
        # Or, make sure we did not break setting user's power levels in a normal flow
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 10
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

    def test_non_admin_cannot_update_power_levels(self) -> None:
        """A non-admin user in the room should not be able to update power levels."""
        room_id = self._create_managed_room()

        # Invite and join the non-admin to the room
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
        )
        self.helper.join(room_id, user=self.non_admin, tok=self.non_admin_token)

        pl = self._get_power_levels(room_id)
        pl["users"][self.non_admin] = 10
        self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            pl,
            self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_non_managed_room_not_affected(self) -> None:
        """Power level changes in non-managed rooms should not be affected by the hook."""
        room_id = self.helper.create_room_as(
            self.creator, tok=self.creator_access_token, is_public=False
        )
        pl = self.helper.get_state(
            room_id, EventTypes.PowerLevels, self.creator_access_token
        )

        pl["users"][self.non_admin] = pl["users"].get(self.creator, 100)
        self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            pl,
            self.creator_access_token,
            expect_code=HTTPStatus.OK,
        )



from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from parameterized import parameterized, parameterized_class
from synapse import event_auth
from synapse.api.constants import (
    CREATOR_POWER_LEVEL,
    EventTypes,
    GuestAccess,
    Membership,
)
from synapse.api.errors import Codes, SynapseError
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS
from synapse.server import HomeServer
from synapse.types import JsonDict, StateMap
from synapse.types.state import StateFilter
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.client import FamedlyControlError
from famedly_control_synapse.rest.room import MANAGED_ROOM_TYPE
from famedly_control_synapse.rest.types import CreateManagedRoomRequest, CreationContent
from famedly_control_synapse.room_handler import famedly_control_user_sync_error
from tests.utils.homeserver_testcase import override_config
from tests.utils.module_api_testcase import ModuleApiTestCase

CREATE_KEY = (EventTypes.Create, "")
POWER_KEY = (EventTypes.PowerLevels, "")
GUEST_ACCESS_KEY = (EventTypes.GuestAccess, "")
JOIN_RULES_KEY = (EventTypes.JoinRules, "")


@parameterized_class(("room_version",), [("10",), ("12",)])
@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
class TestManagedRoomCreation(ModuleApiTestCase):
    room_version: str

    def room_config(self):
        config = CreateManagedRoomRequest(
            room_alias_name="test_room_alias",
            name="Test Room",
            room_version=self.room_version,
            topic="This is a test room",
            groups=["test_group"],
        )
        return config.model_dump()

    def invalid_room_config(self):
        config = {
            "room_alias_name": "test_room_alias",
            "name": "Test Room",
            "topic": "This is a test room",
            "3pid_invites": ["something"],  # Invalid field
        }
        return config

    def _get_creator_powerlevel(self) -> int:
        """
        Per the room version, what is our defined room creator power level
        """
        if KNOWN_ROOM_VERSIONS[self.room_version].msc4289_creator_power_enabled:
            # Room's v12 and up have a power level that is not representable in
            # canonicaljson, 2**53, which is 9007199254740992
            return CREATOR_POWER_LEVEL
        # Rooms prior to version 12 are limited to (2**53) - 1, which is representable
        # and therefore can be in the JSON structure
        return CREATOR_POWER_LEVEL - 1

    def _get_state_map_of_room(self, room_id: str) -> StateMap:
        return self.get_success(
            self.hs.get_storage_controllers().state.get_current_state(
                room_id,
                StateFilter.from_types(
                    [
                        POWER_KEY,
                        CREATE_KEY,
                        GUEST_ACCESS_KEY,
                        JOIN_RULES_KEY,
                    ]
                ),
            )
        )

    def test_room_creation_success(self, mock_get_group_members) -> None:
        """Tests that managed room creation returns the expected response"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.room_config(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert "room_id" in channel.json_body, "Response should contain room_id"
        room_id = channel.json_body["room_id"]

        # Check if the room has the correct configuration
        state_map = self._get_state_map_of_room(room_id)

        assert state_map[EventTypes.JoinRules, ""].content == {
            "join_rule": "invite"
        }, "Join rules should be set to invite"
        assert state_map[EventTypes.GuestAccess, ""].content == {
            "guest_access": "forbidden"
        }, "Guest access should be set to forbidden"

        # Check if account data is updated
        account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert account_data == {MANAGED_ROOM_TYPE: {"groups": ["test_group"]}}

    def test_room_creation_invalid_body(self, mock_get_group_members) -> None:
        """Tests that invalid request body returns a 400 error with an error message"""
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.invalid_room_config(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

    def test_room_creation_with_missing_required_fields(
        self, mock_get_group_members
    ) -> None:
        """Tests that missing required fields in the request body returns a 400 error with an error message"""
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content={"name": "Test Room"},
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

    def test_default_room_powerlevels(self, mock_get_group_members) -> None:
        """Test that a room's default powerlevels are as expected"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        room_config = self.room_config()
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        room_id = channel.json_body["room_id"]

        # Check if the creator has infinite power level and no other user has the same power level
        state_map = self._get_state_map_of_room(room_id)

        creator_pl = event_auth.get_user_power_level(self.creator, state_map)
        assert creator_pl == self._get_creator_powerlevel()

        # Check the invited user's power level
        invitee_pl = event_auth.get_user_power_level(self.invitee, state_map)
        assert invitee_pl == 0, "Invitee should have power level 0"

    def test_user_powerlevel_override_not_destroyed(
        self, mock_get_group_members
    ) -> None:
        """Test that a user powerlevel override is honored by room creation request"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        room_config = self.room_config()
        power_level_content_override = room_config.setdefault(
            "power_level_content_override", {}
        )
        users = power_level_content_override.setdefault("users", {})
        users.setdefault(self.invitee, 1)
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        room_id = channel.json_body["room_id"]

        # Check if the creator has ultimate power level and no other user has the same power level
        state_map = self._get_state_map_of_room(room_id)

        creator_pl = event_auth.get_user_power_level(self.creator, state_map)
        assert creator_pl == self._get_creator_powerlevel()

        # Check the invited user's power level
        invitee_pl = event_auth.get_user_power_level(self.invitee, state_map)
        assert invitee_pl == 1, "Invitee should have power level 1"

    def test_user_powerlevel_override_at_critical_level_forbidden(
        self, mock_get_group_members
    ) -> None:
        """Test that a user powerlevel override is not honored when set too high"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        room_config = self.room_config()
        power_level_content_override = room_config.setdefault(
            "power_level_content_override", {}
        )
        users = power_level_content_override.setdefault("users", {})
        users.setdefault(self.invitee, CREATOR_POWER_LEVEL - 1)
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        # Should error with a 400 Bad Request
        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result

    @parameterized.expand([("invite",), ("ban",), ("kick",)])
    def test_membership_action_powerlevel_cannot_be_circumvented(
        self, mock_get_group_members, membership_action: str
    ) -> None:
        """Test that membership action powerlevels can not be overridden"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids

        room_config = self.room_config()
        assert "power_level_content_override" in room_config
        # Adjust the specific action level we want to test. This is not supposed to be
        # allowed to change
        room_config["power_level_content_override"][membership_action] = 10

        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        # Should error with 400 Bad Request
        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result

    @parameterized.expand(
        [(EventTypes.PowerLevels,), (EventTypes.JoinRules,), (EventTypes.GuestAccess,)]
    )
    def test_event_type_powerlevel_cannot_be_circumvented(
        self, mock_get_group_members, event_type: str
    ) -> None:
        """Test that event type powerlevels can not be overridden"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids

        room_config = self.room_config()
        assert "power_level_content_override" in room_config
        # Adjust the specific action level we want to test. This is not supposed to be
        # allowed to change
        assert "events" in room_config["power_level_content_override"]
        room_config["power_level_content_override"]["events"][event_type] = 10

        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        # Should be a 400 Bad Request
        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result

    def test_room_created_with_members_joined(self, mock_get_group_members) -> None:
        """Tests that the users of the groups are joined to the room after creation"""
        test_group = "test_group_1"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        group_members = [test_member_1, test_member_2]
        mock_get_group_members.return_value = (
            group_members  # in real case this should be external_ids
        )

        # Create a managed room with a group that has members
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[test_group]
        )

        # Check if the member of the group is joined to the room
        for member in group_members:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert channel.json_body["membership"] == "join", channel.json_body[
                "membership"
            ]

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_group]}
        }, room_account_data

    @override_config({"run_background_tasks_on": "made_worker_name"})
    def test_non_background_worker_managed_room_creation_endpoint_disabled(
        self, mock_get_group_members
    ) -> None:
        """
        Tests that the managed room creation endpoint does not function on the incorrect
        Synapse worker
        """
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.room_config(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, channel.result

    def room_config_custom(
        self,
        creation_content: JsonDict | None = None,
        initial_state: list[JsonDict] | None = None,
    ) -> JsonDict:
        config_model = CreateManagedRoomRequest(
            room_alias_name="test_room_alias",
            name="Test Room",
            room_version=self.room_version,
            topic="This is a test room",
            groups=["test_group"],
        )
        config = config_model.model_dump()

        creation_content_dumped = CreationContent().model_dump()
        if creation_content:
            creation_content_dumped.update(**creation_content)

        if initial_state:
            config["initial_state"] = initial_state

        config["creation_content"] = creation_content_dumped
        return config

    def test_room_creation_fails_with_invalid_initial_state(
        self, mock_get_group_members
    ) -> None:
        """
        Test edge cases involving `initial_state` included in the room creation request.
        Specifically do not use the Pydantic models for this, as the model will not pass
        validation during construction. This is to test the handling of that failed
        validation by the endpoint.
        """

        custom_room_config = self.room_config_custom(
            initial_state=[
                {
                    "type": EventTypes.JoinRules,
                    "state_key": "",
                    "content": {"join_rule": Membership.JOIN},
                },
            ],
        )

        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=custom_room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

        custom_room_config = self.room_config_custom(
            initial_state=[
                {
                    "type": EventTypes.GuestAccess,
                    "state_key": "",
                    "content": {"guest_access": GuestAccess.CAN_JOIN},
                },
            ],
        )

        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=custom_room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

        custom_room_config = self.room_config_custom(
            initial_state=[
                {
                    "type": EventTypes.PowerLevels,
                    "state_key": "",
                    "content": {"invite": 10},
                },
            ],
        )

        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=custom_room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )

        # Passing in a power level event to initial_state with invalid content should
        # not be allowed
        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result


@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
class TestAssignGroupsToManagedRoom(ModuleApiTestCase):

    def test_update_single_group_to_single_group(self, mock_get_group_members) -> None:
        """Tests that managed room which already have groups can be updated and the
        account data is updated correctly."""
        test_old_group = "test_old_group"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        old_group_members = [test_member_1, test_member_2]

        test_new_group = "test_new_group"
        test_member_3 = self.register_user("test_member_3", "password")
        new_group_members = [test_member_2, test_member_3]

        def get_group_members(group_id):  # in real case this should return external_ids
            if group_id == test_old_group:
                return old_group_members
            elif group_id == test_new_group:
                return new_group_members
            return []

        mock_get_group_members.side_effect = get_group_members

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[test_old_group]
        )
        self._test_get_membership(room_id, old_group_members, expect_code=200)
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_old_group]}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": [test_new_group]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Check if the new member of the group is joined to the room
        for member in new_group_members:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        # Check if the old member who is not in the new group is removed from the room
        path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{test_member_1}"
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert (
            channel.code == HTTPStatus.OK
        ), f"Expected 200 but got {channel.code} for member {test_member_1}"
        assert (
            channel.json_body["membership"] == "leave"
        ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {test_member_1}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_new_group]}
        }, room_account_data

    def test_update_single_group_to_multiple_groups(
        self, mock_get_group_members
    ) -> None:
        """
        Test updating a single group to multiple groups in a managed room.
        Old groups info: test_group_1
        New groups info: test_group_1, test_group_2, test_group_3
        Edge case to consider with member r who was in group 1 but removed from group later.
        """
        test_group_1 = "test_group_1"
        test_member_a = self.register_user("test_member_a", "password")
        test_member_b = self.register_user("test_member_b", "password")
        test_member_r = self.register_user("test_member_r", "password")
        test_group_1_current_members = [test_member_a, test_member_b]
        test_group_1_creation_members = [test_member_a, test_member_b, test_member_r]

        test_group_2 = "test_group_2"
        test_member_c = self.register_user("test_member_c", "password")
        test_group_2_members = [test_member_b, test_member_c]

        test_group_3 = "test_group_3"
        test_member_d = self.register_user("test_member_d", "password")
        test_member_e = self.register_user("test_member_e", "password")
        test_group_3_members = [test_member_d, test_member_e]

        old_group_info = [test_group_1]
        new_group_info = [test_group_1, test_group_2, test_group_3]

        call_counts = {}

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id not in call_counts:
                call_counts[group_id] = 0
            # Track how many times each group_id has been called
            call_counts[group_id] += 1

            if group_id == test_group_1:
                # First call returns creation members, subsequent calls return current members
                if call_counts[group_id] == 1:
                    return test_group_1_creation_members
                else:
                    return test_group_1_current_members
            elif group_id == test_group_2:
                return test_group_2_members
            elif group_id == test_group_3:
                return test_group_3_members
            return []

        mock_get_group_members.side_effect = get_members_by_group

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=old_group_info
        )

        self._test_get_membership(
            room_id, test_group_1_creation_members, expect_code=200
        )
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": old_group_info}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": new_group_info},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        member_should_be_in_room = set(
            test_group_1_current_members + test_group_2_members + test_group_3_members
        )
        member_should_not_be_in_room = [test_member_r]

        # Check if the new member of the group is joined to the room
        for member in member_should_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        for member in member_should_not_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "leave"
            ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {member}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": new_group_info}
        }, room_account_data

    def test_update_multiple_groups_to_multiple_groups(
        self, mock_get_group_members
    ) -> None:
        """
        Test updating multiple groups to multiple groups in a managed room.
        Old groups info: test_group_1, test_group_2, test_group_3
        New groups info: test_group_1, test_group_4
        The edge cases to consider with member_r and member_b who are in multiple groups.
        """
        test_group_1 = "test_group_1"
        test_member_a = self.register_user("test_member_a", "password")
        test_member_b = self.register_user("test_member_b", "password")
        test_member_r = self.register_user("test_member_r", "password")
        test_group_1_current_members = [test_member_a, test_member_b]
        test_group_1_creation_members = [test_member_a, test_member_b, test_member_r]

        test_group_2 = "test_group_2"
        # Edge case example: member b is in both group 1 and 2. Group 1 is staying in
        # the room and group 2 is removed from the room.
        # member b should stay
        test_member_c = self.register_user("test_member_c", "password")
        test_group_2_members = [test_member_b, test_member_c]

        test_group_3 = "test_group_3"
        test_member_d = self.register_user("test_member_d", "password")
        test_member_e = self.register_user("test_member_e", "password")
        test_group_3_members = [test_member_d, test_member_e]

        test_group_4 = "test_group_4"
        # Edge case example: member r is removed from group 1, but added by group 4
        # and the group 1 remains to the room and group 4 is added to the room.
        # member r should stay
        test_member_f = self.register_user("test_member_f", "password")
        test_group_4_members = [test_member_f, test_member_r]

        old_group_info = [test_group_1, test_group_2, test_group_3]
        new_group_info = [test_group_1, test_group_4]

        call_counts = {}

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id not in call_counts:
                call_counts[group_id] = 0
            call_counts[group_id] += 1

            if group_id == test_group_1:
                # First call returns creation members, subsequent calls return current members
                if call_counts[group_id] == 1:
                    return test_group_1_creation_members
                else:
                    return test_group_1_current_members
            elif group_id == test_group_2:
                return test_group_2_members
            elif group_id == test_group_3:
                return test_group_3_members
            elif group_id == test_group_4:
                return test_group_4_members
            return []

        mock_get_group_members.side_effect = get_members_by_group

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=old_group_info
        )
        self._test_get_membership(
            room_id,
            test_group_1_creation_members + test_group_2_members + test_group_3_members,
            expect_code=200,
        )
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": old_group_info}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": new_group_info},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        member_should_be_in_room = test_group_1_current_members + test_group_4_members
        member_should_not_be_in_room = [test_member_c] + test_group_3_members

        # Check if the new member of the group is joined to the room
        for member in member_should_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        for member in member_should_not_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "leave"
            ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {member}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": new_group_info}
        }, room_account_data

    def test_prevent_room_creator_membership_change(
        self, mock_get_group_members
    ) -> None:
        """Tests that the creator's membership is not changed even if the creator is in the group and the group is removed from the room."""
        group_including_creator = "test_group_including_creator"
        member_1 = self.register_user("test_member_1", "password")
        member_2 = self.register_user("test_member_2", "password")
        group_including_creator_members = [self.creator, member_1, member_2]

        new_group = "new_group"
        member_3 = self.register_user("test_member_3", "password")
        new_group_members = [member_3]

        def get_group_members(group_id):  # in real case this should return external_ids
            if group_id == group_including_creator:
                return group_including_creator_members
            elif group_id == new_group:
                return new_group_members
            return []

        mock_get_group_members.side_effect = get_group_members

        # Create a managed room with a group that has the creator as a member
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[group_including_creator]
        )

        # Now update the group information to remove the group from the room
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": [new_group]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Check if the creator is still joined to the room
        path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{self.creator}"
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert (
            channel.code == HTTPStatus.OK
        ), f"Expected 200 but got {channel.code} for member {self.creator}"
        assert (
            channel.json_body["membership"] == "join"
        ), f"Expected membership to be join but got {channel.json_body['membership']} for member {self.creator}"

    def test_client_error_response(self, mock_get_group_members) -> None:
        """Test that if client raises an error, the endpoint returns an error with an
        error message."""
        # Create a managed room
        test_group = "test_group"
        test_member = self.register_user("test_member", "password")
        mock_get_group_members.return_value = [test_member]
        room_id = self._create_managed_room(name="Test Room", groups=[test_group])

        # Try to update groups with the client raising an error
        mock_get_group_members.side_effect = FamedlyControlError(
            HTTPStatus.INTERNAL_SERVER_ERROR, "Service unavailable"
        )
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": ["new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == 500, channel.result
        assert "Service unavailable" in channel.json_body["error"]

    def test_conversion_error_response(self, mock_get_group_members) -> None:
        """Test that if batch conversion raises an error, the endpoint returns an error
        with an error message."""
        # Create a managed room
        test_group = "test_group"
        test_member = self.register_user("test_member", "password")
        mock_get_group_members.return_value = [test_member]
        room_id = self._create_managed_room(name="Test Room", groups=[test_group])

        # Try to update groups and
        # batch_convert_external_user_ids_to_matrix_user_ids will return not_founds. The
        # call to the external api for getting the members of group "new_group" need to
        # not exist in the local database, which raises the error
        mock_get_group_members.return_value = ["external_id_1", "external_id_2"]
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": ["new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Some external user IDs could not be mapped to Matrix user IDs"
            in channel.json_body["error"]
        )
        assert sorted(channel.json_body["details"]) == [
            "external_id_1",
            "external_id_2",
        ]

    def test_leave_error_all_fail_response(
        self,
        mock_get_group_members,
    ) -> None:
        """Test that if leaving room raises errors for different users with different
        error codes, the endpoint returns the details and metrics are incremented for
        each error code.
        """
        # Get the metric initial values for both error codes
        server_name = self.hs.hostname
        initial_value_unknown = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=server_name
        )._value.get()
        initial_value_unauthorized = famedly_control_user_sync_error.labels(
            error_code="M_UNAUTHORIZED", server_name=server_name
        )._value.get()

        # Create a managed room with test_group_1
        test_group_1 = "test_group_1"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        mock_get_group_members.return_value = [test_member_1, test_member_2]
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Prepare test_group_2 which removes all 2 members and add a new member
        test_group_2 = "test_group_2"
        test_member_3 = self.register_user("test_member_3", "password")
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [test_member_3]
        )

        # Try to update groups with update_room_membership raising errors for different users
        def update_room_membership_side_effect(
            sender, target, room_id, new_membership, **kwargs
        ):
            if target == test_member_1:
                raise SynapseError(
                    HTTPStatus.INTERNAL_SERVER_ERROR, "Unexpected error", Codes.UNKNOWN
                )
            elif target == test_member_2:
                raise SynapseError(
                    HTTPStatus.UNAUTHORIZED, "Permission denied", Codes.UNAUTHORIZED
                )
            return None

        with patch(
            "synapse.module_api.ModuleApi.update_room_membership",
            new_callable=AsyncMock,
        ) as mock_update:
            mock_update.side_effect = update_room_membership_side_effect

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to remove some members from the room"
                in channel.json_body["error"]
            )
            assert test_member_1 in channel.json_body["details"]
            assert test_member_2 in channel.json_body["details"]

        # Since both members failed to be removed, they should still be in the room
        # and the new member should be joined to the room
        for member in [test_member_1, test_member_2, test_member_3]:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        # Check both metrics are incremented by 1 each
        new_value_unknown = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=server_name
        )._value.get()
        new_value_unauthorized = famedly_control_user_sync_error.labels(
            error_code="M_UNAUTHORIZED", server_name=server_name
        )._value.get()
        assert new_value_unknown == initial_value_unknown + 1
        assert new_value_unauthorized == initial_value_unauthorized + 1

    def test_join_error_response(
        self,
        mock_get_group_members,
    ) -> None:
        """Test that if joining room raises an error, the endpoint returns an error with
        an error message and error metric was sent."""
        # Get the metric initial value
        server_name = self.hs.hostname
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=server_name
        )._value.get()

        # Create a managed room with test_group_1 which has 2 valid members.
        test_group_1 = "test_group_1"
        test_group_2 = "test_group_2"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        non_existent_user = f"@random_user:{server_name}"
        # Go ahead and register the external user id, this makes
        # `batch_convert_external_user_ids_to_matrix_user_ids()` happy so we can check
        # the response from `force_join_users_to_room()`. How realistic this is: unknown
        self.register_external_id(non_existent_user)
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [test_member_2, non_existent_user]
        )
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Now update the group assignment to test_group_2, which has an already-joined
        # member and one non-existent member. `force_join_users_to_room` will skip for
        # the already-joined member and return error record for the non-existent member.
        def update_room_membership_side_effect(
            sender, target, room_id, new_membership, **kwargs
        ):
            if target == non_existent_user:
                raise SynapseError(404, "User not found", Codes.NOT_FOUND)
            return None

        with patch(
            "synapse.module_api.ModuleApi.update_room_membership",
            new_callable=AsyncMock,
        ) as mock_update:
            mock_update.side_effect = update_room_membership_side_effect

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to add some members to the room" in channel.json_body["error"]
            )
            assert non_existent_user in channel.json_body["details"]
            # test_member_2 should not be in details as the member was already in the room
            assert test_member_2 not in channel.json_body["details"]

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_both_join_and_leave_error_response(
        self,
        mock_get_group_members,
    ) -> None:
        """Test that if both joining and leaving room raise errors, the endpoint returns
        an error message with both leave_errors and join_errors."""
        # Create a managed room with test_group_1
        test_group_1 = "test_group_1"
        test_group_2 = "test_group_2"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        non_existent_user = f"@random_user:{self.hs.hostname}"
        # Go ahead and register the external user id, this makes
        # `batch_convert_external_user_ids_to_matrix_user_ids()` happy so we can check
        # the response from `force_join_users_to_room()`. How realistic this is: unknown
        self.register_external_id(non_existent_user)
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [non_existent_user]
        )
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Now update the group assignment to test_group_2, which only has non-existent member.
        # Both `force_join_users_to_room` and `remove_users_from_room` will return errors.
        with (
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.remove_users_from_room",
                new_callable=AsyncMock,
            ) as mock_remove,
        ):
            mock_remove.return_value = {test_member_1: "Unexpected error"}

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to add and remove some members from the room"
                in channel.json_body["error"]
            )
            assert "leave_errors" in channel.json_body["details"]
            assert "join_errors" in channel.json_body["details"]
            assert test_member_1 in channel.json_body["details"]["leave_errors"]
            assert non_existent_user in channel.json_body["details"]["join_errors"]

    @override_config({"run_background_tasks_on": "fake_worker"})
    def test_non_background_worker_assign_groups_endpoint_disabled(
        self, mock_get_group_members
    ) -> None:
        """
        Tests that the assign groups endpoint does not function on the incorrect
        Synapse worker
        """
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + "/!fake_room_does_not_matter/groups",
            content={"groups": ["test_new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.NOT_FOUND, channel.result


@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
class TestListManagedRooms(ModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.non_admin = self.register_user("non_admin", "password", admin=False)
        self.non_admin_token = self.login("non_admin", "password")
        self.account_data_handler = homeserver.get_account_data_handler()

    def test_list_requires_admin(self, mock_get_group_members) -> None:
        """Non-admin users should get a 403."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.non_admin_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.FORBIDDEN, channel.result

    def test_list_requires_auth(self, mock_get_group_members) -> None:
        """Unauthenticated requests should get a 401."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.UNAUTHORIZED, channel.result

    def test_list_empty(self, mock_get_group_members) -> None:
        """Listing when no managed rooms exist should return empty chunk."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["chunk"] == []
        assert channel.json_body["total_room_count_estimate"] == 0

    def test_list_returns_managed_rooms(self, mock_get_group_members) -> None:
        """Created managed rooms should appear in the listing."""
        user_1 = self.register_user("user1", "password")
        user_2 = self.register_user("user2", "password")

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id == "group1":
                return [user_1]
            if group_id == "group2":
                return [user_2]
            return []

        mock_get_group_members.side_effect = get_members_by_group

        room_id = self._create_managed_room(
            name="Listed Room", groups=["group1", "group2"]
        )

        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["total_room_count_estimate"] == 1
        assert len(channel.json_body["chunk"]) == 1

        room_chunk = channel.json_body["chunk"][0]
        assert room_chunk["room_id"] == room_id
        assert room_chunk["de.famedly.managedRoom"]["groups"] == [
            "group1",
            "group2",
        ]

    def test_list_pagination(self, mock_get_group_members) -> None:
        """Pagination should work with from and limit params."""
        user_1 = self.register_user("user1", "password")
        user_2 = self.register_user("user2", "password")
        user_3 = self.register_user("user3", "password")

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id == "group1":
                return [user_1]
            if group_id == "group2":
                return [user_2]
            if group_id == "group3":
                return [user_3]
            return []

        mock_get_group_members.side_effect = get_members_by_group

        room_ids = []
        for i in range(3):
            room_ids.append(
                self._create_managed_room(name=f"Room {i}", groups=[f"group{i}"])
            )

        # Get first page with limit 2
        channel = self.make_request(
            method="GET",
            path=f"{self.LIST_PATH}?limit=2",
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert len(channel.json_body["chunk"]) == 2
        assert channel.json_body["total_room_count_estimate"] == 3
        assert "next_batch" in channel.json_body
        assert "prev_batch" not in channel.json_body

        # Get second page
        next_batch = channel.json_body["next_batch"]
        channel = self.make_request(
            method="GET",
            path=f"{self.LIST_PATH}?from={next_batch}&limit=2",
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert len(channel.json_body["chunk"]) == 1
        assert "next_batch" not in channel.json_body
        assert "prev_batch" in channel.json_body

    @override_config({"run_background_tasks_on": "fake_worker"})
    def test_non_background_worker_list_endpoint_disabled(
        self, mock_get_group_members
    ) -> None:
        """
        Tests that the assign groups endpoint does not function on the incorrect
        Synapse worker
        """
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.NOT_FOUND, channel.result



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from unittest.mock import AsyncMock, MagicMock

from parameterized import parameterized
from synapse.api.errors import HttpResponseException
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.defer import ensureDeferred
from twisted.internet.testing import MemoryReactor
from twisted.trial import unittest

from famedly_control_synapse.client import FamedlyControlClient, FamedlyControlError
from famedly_control_synapse.config import FamedlyControlConfig
from tests.utils.module_api_testcase import ModuleApiTestCase


def _make_client(api_url: str):
    config = FamedlyControlConfig(
        famedly_control={
            "api_url": api_url,
            "access_token": "test_token",
        },
        auth_provider="https://idp.example.com/",
    )
    api = MagicMock()
    api.http_client = MagicMock()
    api.http_client.post_json_get_json = AsyncMock()
    return api, FamedlyControlClient(api, config)


class TestClientUrlConstruction(unittest.TestCase):
    """Ensure the client builds endpoint URLs without double slashes."""

    api_urls_with_path = [
        "http://api.example.com/famedlyControl",
        "http://api.example.com/famedlyControl/",
    ]

    api_urls_bare_domain = [
        "http://api.example.com",
        "http://api.example.com/",
    ]

    def test_get_group_members_url_with_path(self):
        for api_url in self.api_urls_with_path:
            api, client = _make_client(api_url)
            api.http_client.post_json_get_json.return_value = {"Ok": {"members": []}}
            self.successResultOf(ensureDeferred(client.get_group_members("some_group")))

            called_uri = api.http_client.post_json_get_json.call_args[0][0]
            self.assertEqual(
                called_uri,
                "http://api.example.com/famedlyControl/get_group_members",
                f"Failed for api_url={api_url!r}",
            )

    def test_get_group_members_url_bare_domain(self):
        for api_url in self.api_urls_bare_domain:
            api, client = _make_client(api_url)
            api.http_client.post_json_get_json.return_value = {"Ok": {"members": []}}
            self.successResultOf(ensureDeferred(client.get_group_members("some_group")))

            called_uri = api.http_client.post_json_get_json.call_args[0][0]
            self.assertEqual(
                called_uri,
                "http://api.example.com/get_group_members",
                f"Failed for api_url={api_url!r}",
            )

    def test_get_all_groups_diffs_url_with_path(self):
        for api_url in self.api_urls_with_path:
            api, client = _make_client(api_url)
            api.http_client.post_json_get_json.return_value = {
                "Ok": {"next_sync": "1", "data": {}}
            }
            self.successResultOf(ensureDeferred(client.get_all_groups_diffs(sync=None)))

            called_uri = api.http_client.post_json_get_json.call_args[0][0]
            self.assertEqual(
                called_uri,
                "http://api.example.com/famedlyControl/get_all_groups_diffs",
                f"Failed for api_url={api_url!r}",
            )

    def test_get_all_groups_diffs_url_bare_domain(self):
        for api_url in self.api_urls_bare_domain:
            api, client = _make_client(api_url)
            api.http_client.post_json_get_json.return_value = {
                "Ok": {"next_sync": "1", "data": {}}
            }
            self.successResultOf(ensureDeferred(client.get_all_groups_diffs(sync=None)))

            called_uri = api.http_client.post_json_get_json.call_args[0][0]
            self.assertEqual(
                called_uri,
                "http://api.example.com/get_all_groups_diffs",
                f"Failed for api_url={api_url!r}",
            )


class TestClientResponse(ModuleApiTestCase):

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.client = self.hs.room_control.client

    def test_request_success(self):
        """Test that client returns the expected list of member IDs on success."""
        expected_members = ["user1_external_id", "user2_external_id"]

        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={
                "Ok": {
                    "members": [
                        {"user_id": "user1_external_id"},
                        {"user_id": "user2_external_id"},
                    ]
                }
            }
        )
        members = self.get_success(self.client.get_group_members("test_group"))
        assert members == expected_members

    @parameterized.expand([("Forbidden", 403), ("Unauthorized", 401)])
    def test_request_fail_with_err_response_types(self, error_type, expected_code):
        """Test that client returns 200 with Err message raises FamedlyControlError with proper code."""
        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={"Err": {"type": error_type}}
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == expected_code
        assert (
            failure.value.msg == f"Famedly Control API: Error in response: {error_type}"
        )

    def test_request_fail_with_err_response_unknown_type(self):
        """Test that client returns 200 with Err message with unknown type raises FamedlyControlError with 500 code."""
        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={"Err": {"type": "SomeUnknownType"}}
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        self.assertEqual(failure.value.code, 500)
        assert failure.value.code == 500
        assert (
            failure.value.msg
            == "Famedly Control API: Error in response: SomeUnknownType"
        )

    def test_request_fail_with_err_response_not_dict(self):
        """Test that client returns 200 with Err message with non dict format raises FamedlyControlError with 500 code."""
        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={"Err": "SomeError"}
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == 500
        assert (
            failure.value.msg
            == "Famedly Control API: Unexpected error: 'str' object has no attribute 'get'"
        )

    def test_request_fail_with_unexpected_response(self):
        """Test that client raises FamedlyControlError when response format is neither "Ok" nor "Err"."""
        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={"Unexpected": "format"}
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == 502
        assert (
            failure.value.msg
            == "Famedly Control API: Unexpected response format: {'Unexpected': 'format'}"
        )

    def test_request_fail_with_http_exception(self):
        """Test that client raises any other HTTP error raises FamedlyControlError."""
        self.client.http_client.post_json_get_json = AsyncMock(
            side_effect=HttpResponseException(500, "Internal Server Error", b"")
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == 500
        assert (
            failure.value.msg
            == "Famedly Control API: HTTP response error: Internal Server Error"
        )

    def test_request_fail_with_validation_error(self):
        """Test that client raises FamedlyControlError on validation error."""
        self.client.http_client.post_json_get_json = AsyncMock(
            return_value={
                "Ok": {"members": "not_a_list"}
            }  # invalid: members must be a list
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == 500
        assert failure.value.msg.startswith(
            "Famedly Control API: Unexpected error: 1 validation error for GroupMembersResponse"
        )

    def test_request_fail_with_generic_exception(self):
        """Test that client raises FamedlyControlError on any unexpected exception."""
        self.client.http_client.post_json_get_json = AsyncMock(
            side_effect=RuntimeError("something broke")
        )
        failure = self.get_failure(
            self.client.get_group_members("test_group"), FamedlyControlError
        )
        assert failure.value.code == 500
        assert (
            failure.value.msg
            == "Famedly Control API: Unexpected error: something broke"
        )



from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from pydantic.fields import FieldInfo
from pydantic_core.core_schema import ValidationInfo
from synapse.api.constants import (
    CREATOR_POWER_LEVEL,
    EventTypes,
    GuestAccess,
    Membership,
)
from synapse.types import JsonDict
from typing_extensions import Self

MANAGED_ROOM_TYPE = "de.famedly.managedRoom"
SYNC_TOKEN_TYPE = "de.famedly.roomControl.lastSyncToken.v1"


class CreationContent(BaseModel):
    """Pydantic model for CreationContent."""

    creator: str | None = None
    m_federate: Literal[False] = Field(
        default=False,
        alias="m.federate",
        description="Federation is disabled for managed rooms.",
    )
    predecessor: JsonDict | None = None
    room_version: str | None = (
        None  # The room version here is overridden. It is irrelevant which version comes here.
    )
    type: str | None = None

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


PROTECTED_EVENT_TYPES = {
    EventTypes.PowerLevels: CREATOR_POWER_LEVEL - 1,
    EventTypes.JoinRules: CREATOR_POWER_LEVEL - 1,
    EventTypes.GuestAccess: CREATOR_POWER_LEVEL - 1,
}


class PowerLevelEventContent(BaseModel):
    """Power level event content for overriding default power levels."""

    # A full model validator is below to verify that membership action power levels can
    # not be overridden
    ban: int = CREATOR_POWER_LEVEL - 1
    # events has a validator below
    events: dict[str, int] = Field(
        default_factory=lambda: {
            EventTypes.PowerLevels: CREATOR_POWER_LEVEL - 1,
            EventTypes.JoinRules: CREATOR_POWER_LEVEL - 1,
            EventTypes.GuestAccess: CREATOR_POWER_LEVEL - 1,
        }
    )
    # Use the `exclude_if` pattern for ensuring that this field is not included if it is
    # the default per the matrix spec. This allows these defaults to not be included
    # when serializing this model. Other options included having `exclude_unset` and
    # `exclude_defaults` from `model_dump()`, but that would mean the fields we override
    # to also be excluded and that is unwanted.
    events_default: int = Field(0, exclude_if=lambda x: x == 0)
    invite: int = CREATOR_POWER_LEVEL - 1
    kick: int = CREATOR_POWER_LEVEL - 1
    redact: int = Field(50, exclude_if=lambda x: x == 50)
    state_default: int = Field(50, exclude_if=lambda x: x == 50)
    # users has a validator below
    users: dict[str, int] = Field(default_factory=dict)
    users_default: int = Field(0, exclude_if=lambda x: x == 0)
    notifications: dict[str, int] = Field(default_factory=dict)

    @field_validator("events")
    @classmethod
    def validate_events(cls, v: dict[str, int], info: ValidationInfo) -> dict[str, int]:
        """
        Ensure that certain event types can not have their power level overridden
        """
        for event_type, power_level in v.items():
            # First check for the specific event types that are not allowed to be
            # changed away from the room creator
            if event_type in PROTECTED_EVENT_TYPES:
                immutable_power_level = PROTECTED_EVENT_TYPES.get(event_type)
                if power_level != immutable_power_level:
                    # This will return a 400 on the room creation endpoint
                    raise ValueError(
                        f"Changing power_level of '{event_type}' in {info.field_name} is forbidden!"
                    )
        return v

    @field_validator("users")
    @classmethod
    def validate_users(cls, v: dict[str, int], info: ValidationInfo) -> dict[str, int]:
        """
        Ensure that any user that is not the room creator is not allowed to have the
        same power level as the room creator
        """
        # The context object is of type ContextT(and appears to be a proxy object type),
        # but only if it was passed into the model validation.
        if not isinstance(info.context, dict):
            raise ValueError(
                "Context should be passed into the model validation in the form of a dict"
            )

        room_creator: str | None = info.context.get("room_creator")
        if not room_creator:
            raise ValueError(
                "Room creator key was found in context, but not a usable value"
            )

        for user, power_level in v.items():
            if user != room_creator and power_level == CREATOR_POWER_LEVEL - 1:
                raise ValueError(
                    "Can not have a user with that high a power level, only the room creator"
                )
            if user == room_creator and power_level != CREATOR_POWER_LEVEL - 1:
                raise ValueError("Can not change the room creator's power level")
        return v

    @model_validator(mode="before")
    @classmethod
    def merge_events_overrides(cls, data: Any) -> Any:
        """
        Allow merging new 'events' overrides into the expected data. This will not do
        validation
        """
        if isinstance(data, dict):
            # Retrieve the defaults from the default_factory on the 'events' object
            event_field: FieldInfo = cls.model_fields["events"]

            # mypy thinks that default_factory() here has both too few arguments and
            # that None is not callable. Neither of these is true
            default_events_items = event_field.default_factory()  # type: ignore[misc, call-arg]

            if "events" in data and isinstance(data["events"], dict):
                # Update the defaults with those passed in on creation of this object
                default_events_items.update(data["events"])
                # Overwrite those continuing on into the object creation
                data["events"] = default_events_items

            # if there was no 'events' being asked for, then the normal default_factory
            # will fill in the defaults for us
        return data

    @model_validator(mode="after")
    def validate_membership_actions(self) -> Self:
        """
        Ensure that the membership action power levels can not be overridden
        """
        for action in ("ban", "invite", "kick"):
            action_value = getattr(self, action)
            if action_value != CREATOR_POWER_LEVEL - 1:
                raise ValueError(
                    f"Membership action('{action}') power level can not be overridden"
                )
        return self


class CreateManagedRoomRequest(BaseModel):
    """Request body for creating a managed room."""

    room_alias_name: str
    name: str
    topic: str | None = None
    creation_content: CreationContent = Field(default_factory=CreationContent)
    # initial_state has a validator below
    initial_state: list[JsonDict] = Field(
        default_factory=lambda: [
            {
                "type": EventTypes.GuestAccess,
                "state_key": "",
                "content": {"guest_access": GuestAccess.FORBIDDEN},
            },
            {
                "type": EventTypes.JoinRules,
                "state_key": "",
                "content": {"join_rule": Membership.INVITE},
            },
        ]
    )

    power_level_content_override: PowerLevelEventContent = Field(
        default_factory=PowerLevelEventContent
    )
    room_version: str | None = None
    groups: list[str]
    is_direct: Literal[False] = Field(
        default=False, description="is_direct cannot be enabled for managed rooms."
    )
    visibility: Literal["private"] = Field(
        default="private",
        description="Visibility is fixed to private for managed rooms.",
    )

    model_config = ConfigDict(extra="forbid")

    @field_validator("initial_state")
    @classmethod
    def validate_initial_state(
        cls, v: list[JsonDict], info: ValidationInfo
    ) -> list[JsonDict]:
        for state_dict in v:
            if state_dict.get("type") == EventTypes.JoinRules:
                if state_dict.get("content", {}).get("join_rule") != Membership.INVITE:
                    raise ValueError(
                        f"{info.field_name} contains join_rule that is not 'invite'"
                    )
            if state_dict.get("type") == EventTypes.GuestAccess:
                if (
                    state_dict.get("content", {}).get("guest_access")
                    != GuestAccess.FORBIDDEN
                ):
                    raise ValueError(
                        f"{info.field_name} contains guest_access that is not 'forbidden'"
                    )
            if state_dict.get("type") == EventTypes.PowerLevels:
                PowerLevelEventContent.model_validate(
                    state_dict.get("content"), context=info.context
                )
        return v


class AssignGroupsToManagedRoomRequest(BaseModel):
    """Request body for assigning groups to a managed room."""

    groups: list[str]

    model_config = ConfigDict(extra="forbid")



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Any

from pydantic import ValidationError
from synapse.api.constants import EventTypes, Membership
from synapse.api.errors import Codes, SynapseError
from synapse.events import EventBase
from synapse.http.server import JsonResource
from synapse.module_api import ModuleApi
from synapse.types import StateMap

from famedly_control_synapse.client import FamedlyControlClient
from famedly_control_synapse.config import FamedlyControlConfig
from famedly_control_synapse.repository import ManagedRoomRepository
from famedly_control_synapse.rest.room import (
    MANAGED_ROOM_API_PREFIX,
    AssignGroupsToManagedRoomResource,
    CreateManagedRoomResource,
    ListManagedRoomsResource,
)
from famedly_control_synapse.rest.types import PowerLevelEventContent
from famedly_control_synapse.room_handler import ManagedRoomHandler
from famedly_control_synapse.sync import GroupMembershipSyncer

logger = logging.getLogger(__name__)


class _SyncTriggerJsonResource(JsonResource):
    """JsonResource subclass that attempts to start the background sync
    on every incoming request. If the data isn't ready yet, the next API call
    will retry automatically.
    """

    def __init__(self, hs, syncer: GroupMembershipSyncer):
        super().__init__(hs)
        self._syncer = syncer

    def render(self, request):
        response_body = super().render(request)
        self._syncer.start()
        return response_body


class FamedlyControl:
    __version__ = "0.0.2"

    def __init__(self, config: FamedlyControlConfig, api: ModuleApi) -> None:
        self.api = api
        self.server_name = api.server_name
        self.clock = api._hs.get_clock()
        self.config = config
        self.client = FamedlyControlClient(self.api, config)
        self.room_handler = ManagedRoomHandler(self.api, self.config)
        self.repository = ManagedRoomRepository(api)

        if self.api.should_run_background_tasks():
            self.syncer = GroupMembershipSyncer(
                api, self.client, self.room_handler, self.repository, config
            )

            # Register servlets
            self.resource = _SyncTriggerJsonResource(self.api._hs, self.syncer)
            CreateManagedRoomResource(
                self.api, self.client, self.room_handler, self.repository
            ).register(self.resource)
            ListManagedRoomsResource(self.api, self.repository).register(self.resource)
            AssignGroupsToManagedRoomResource(
                self.api, self.client, self.room_handler, self.repository
            ).register(self.resource)
            self.api.register_web_resource(MANAGED_ROOM_API_PREFIX, self.resource)

            self.api._clock.call_when_running(self.syncer.start)

        self.api.register_third_party_rules_callbacks(
            check_event_allowed=self.check_event_allowed,
        )

        logger.info("Module initialized")

    @staticmethod
    def parse_config(config: dict[str, Any]) -> FamedlyControlConfig:
        return FamedlyControlConfig.model_validate(config)

    async def check_event_allowed(
        self, event: EventBase, state_events: StateMap[EventBase]
    ) -> tuple[bool, dict | None]:
        """Third-party rules callback that enforces membership and power level
        restrictions for managed rooms.
        """
        # Because of the check for a managed room, this function does not run when
        # creating a room. A room can not be marked as managed until after its creation.
        if event.type not in (EventTypes.Member, EventTypes.PowerLevels):
            return True, None

        create_event = state_events.get((EventTypes.Create, ""))
        if create_event is None:
            raise SynapseError(
                500,
                "Managed room is missing m.room.create state",
                Codes.UNKNOWN,
            )

        if not await self.repository.is_managed_room(
            event.room_id, create_event.sender
        ):
            return True, None

        if event.type == EventTypes.Member:
            return await self._check_membership_allowed(event, create_event.sender)
        elif event.type == EventTypes.PowerLevels:
            return await self._check_power_levels_allowed(
                event, state_events, create_event.sender
            )
        else:
            # cannot happen but makes the linter happy
            return True, None

    async def _check_membership_allowed(
        self, event: EventBase, admin_user: str
    ) -> tuple[bool, dict | None]:
        """Block membership changes in managed rooms unless sent by the admin.

        Joins are allowed because managed rooms are invite-only and only
        the admin can send invites, so a join is implicitly admin-authorised.
        """
        if event.sender == admin_user:
            return True, None

        membership = event.content.get("membership")
        if membership == Membership.JOIN:
            return True, None

        raise SynapseError(
            403,
            "Only the admin can edit membership in a managed room",
            Codes.FORBIDDEN,
        )

    async def _check_power_levels_allowed(
        self, event: EventBase, state_events: StateMap[EventBase], room_creator: str
    ) -> tuple[bool, dict | None]:
        """Block invalid power level changes in managed rooms. While any system admin
        can make changes to this event, protect that the levels we consider invariant
        can not be changed. Use the comprehensive Pydantic model built for this purpose.

        Enforced constraints (as a non-exhaustive list):
        * The admin must remain the highest power level in the room.
        * Certain event type power levels are not allowed to be changed
          (``m.room.join_rules``, ``m.room.power_levels``, ``m.room.guest_access``).
        * Membership actions (ban, kick, invite) must not ever be lower than the room
          creator's PL
        """
        new_content = event.content
        try:
            PowerLevelEventContent.model_validate(
                new_content, context={"room_creator": room_creator}
            )
            return True, None
        except ValidationError as e:
            # For returning the error to the client, just select the first error
            err = e.errors()[0]
            single_validation_error = {"loc": err.get("loc"), "msg": err.get("msg")}

            raise SynapseError(
                403,
                f"Invalid request body: {single_validation_error}",
                errcode=Codes.FORBIDDEN,
            )



from unittest import TestCase

import pytest
from parameterized import parameterized
from pydantic import ValidationError
from synapse.api.constants import CREATOR_POWER_LEVEL, EventTypes

from famedly_control_synapse.rest.types import (
    CreateManagedRoomRequest,
    CreationContent,
    PowerLevelEventContent,
)


class TestCreationContent:
    def test_valid_creator_for_v10(self) -> None:
        content = CreationContent(creator="@alice:example.com")
        assert content.creator == "@alice:example.com"

    def test_m_federate_always_false(self) -> None:
        content = CreationContent()
        assert content.m_federate is False

        with pytest.raises(ValidationError):
            content = CreationContent(m_federate=True)


class TestPowerLevelEventContent:
    def test_valid_power_level_event_content(self) -> None:
        json_dict: dict = {
            "users": {"@alice:example.com": 100, "@bob:example.com": 50},
            "users_default": 0,
        }
        plc = PowerLevelEventContent.model_validate(
            json_dict, context={"room_creator": "@room_creator:example.com"}
        )
        assert plc.users["@alice:example.com"] == 100
        assert plc.users["@bob:example.com"] == 50
        assert plc.users_default == 0

    def test_invalid_users_type(self) -> None:
        json_dict: dict = {
            "users": ["@alice:example.com", "@bob:example.com"],
            "users_default": 0,
        }

        with pytest.raises(ValidationError):
            PowerLevelEventContent.model_validate(
                json_dict, context={"room_creator": "@room_creator:example.com"}
            )

    def test_matrix_default_values_are_not_serialized(self) -> None:
        """
        Test verify that serializing the power level event does not contain matrix spec
        defaulted data
        """
        # First test that an empty payload is as expected
        json_dict: dict = {}
        plc = PowerLevelEventContent.model_validate(json_dict)
        assert plc.users_default == 0
        assert plc.ban == CREATOR_POWER_LEVEL - 1
        dumped_dict = plc.model_dump()
        # users_default should be None as 0 is its default
        assert dumped_dict.get("users_default") is None
        # events_default should be None as 0 is its default
        assert dumped_dict.get("events_default") is None
        # redact should be None as 50 is its default
        assert dumped_dict.get("redact") is None
        # state_default should be None as 50 is its default
        assert dumped_dict.get("state_default") is None
        assert dumped_dict.get("ban") == CREATOR_POWER_LEVEL - 1

        # Then test that explicitly giving a value that is the default still is as expected
        json_dict = {
            "users_default": 0,
            "events_default": 0,
            "redact": 50,
            "state_default": 50,
        }
        plc = PowerLevelEventContent.model_validate(json_dict)
        assert plc.users_default == 0
        assert plc.ban == CREATOR_POWER_LEVEL - 1
        dumped_dict = plc.model_dump()
        # users_default should be None as 0 is its default
        assert dumped_dict.get("users_default") is None
        # events_default should be None as 0 is its default
        assert dumped_dict.get("events_default") is None
        # redact should be None as 50 is its default
        assert dumped_dict.get("redact") is None
        # state_default should be None as 50 is its default
        assert dumped_dict.get("state_default") is None
        assert dumped_dict.get("ban") == CREATOR_POWER_LEVEL - 1

        # Then test that making the value a non-default is being produced in the result
        json_dict = {
            "users_default": 1,
            "events_default": 1,
            "redact": 51,
            "state_default": 51,
        }
        plc = PowerLevelEventContent.model_validate(json_dict)
        assert plc.users_default == 1
        assert plc.ban == CREATOR_POWER_LEVEL - 1
        dumped_dict = plc.model_dump()
        # Should all be present as they are explicitly defined as a non-default value
        assert dumped_dict.get("users_default") == 1
        assert dumped_dict.get("events_default") == 1
        assert dumped_dict.get("redact") == 51
        assert dumped_dict.get("state_default") == 51
        assert dumped_dict.get("ban") == CREATOR_POWER_LEVEL - 1


class TestCreateManagedRoomRequest(TestCase):
    """
    Exercise the various options in the shape of the CreateManagedRoomRequest, to make
    sure it behaves as expected.

    As notes:
        * It is not expected here that the creation_content.creator field is filled in, as
        this is overridden by Synapse at the time of room creation.
        * While testing that a room version is passed in and in the correct places, the
        default behavior is to use the default room version defined by the server, so it
        is None here and filled in during the request
    """

    @parameterized.expand([("10",), ("12",)])
    def test_default(self, room_version) -> None:
        """
        Test that a basic object is created with expected parameters
        """
        json_dict = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "room_version": room_version,
            "groups": ["testgroup"],
        }

        req = CreateManagedRoomRequest.model_validate(
            json_dict, context={"room_creator": "@room_creator:example.com"}
        )
        assert req.room_alias_name == "testroom"
        assert req.name == "Test Room"
        assert req.room_version == room_version
        # This does not need to exist here. It will be overridden during the room
        # creation, so is not needed
        assert req.creation_content.room_version is None
        assert req.creation_content.creator is None
        assert req.initial_state == [
            {
                "content": {
                    "guest_access": "forbidden",
                },
                "state_key": "",
                "type": "m.room.guest_access",
            },
            {
                "content": {
                    "join_rule": "invite",
                },
                "state_key": "",
                "type": "m.room.join_rules",
            },
        ]
        assert req.is_direct is False
        assert req.visibility == "private"
        # At the time this object is created, the room creator is not present in room
        # v10, this is added at the time of the request being processed. Users should be
        # empty
        assert len(req.power_level_content_override.users) == 0
        # By default, 3 event types are included in the events object
        self.assertDictEqual(
            req.power_level_content_override.events,
            {
                EventTypes.PowerLevels: CREATOR_POWER_LEVEL - 1,
                EventTypes.JoinRules: CREATOR_POWER_LEVEL - 1,
                EventTypes.GuestAccess: CREATOR_POWER_LEVEL - 1,
            },
        )
        assert req.power_level_content_override.ban == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.invite == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.kick == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.redact == 50
        assert req.power_level_content_override.events_default == 0
        assert req.power_level_content_override.users_default == 0
        assert req.power_level_content_override.state_default == 50

        # A single group was requested, make sure it is there
        assert "testgroup" in req.groups
        assert len(req.groups) == 1

    def test_with_none_powerlevel(self) -> None:
        """
        Test passing a None as the powerlevel override is prohibited
        """
        # Then try by setting as None. Expect a ValidationError
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": None,
            "groups": ["testgroup"],
        }

        with pytest.raises(ValidationError):
            CreateManagedRoomRequest.model_validate(
                room_config, context={"room_creator": "@room_creator:example.com"}
            )

    def test_with_empty_powerlevel(self) -> None:
        """
        Test that adding an empty power level object does not allow circumventing all
        power levels
        """
        # First try with an empty dict, should still get the defaults
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": {},
            "groups": [],
        }

        req = CreateManagedRoomRequest.model_validate(
            room_config, context={"room_creator": "@room_creator:example.com"}
        )
        assert len(req.power_level_content_override.users) == 0
        # By default, 3 event types are included in the events object
        self.assertDictEqual(
            req.power_level_content_override.events,
            {
                EventTypes.PowerLevels: CREATOR_POWER_LEVEL - 1,
                EventTypes.JoinRules: CREATOR_POWER_LEVEL - 1,
                EventTypes.GuestAccess: CREATOR_POWER_LEVEL - 1,
            },
        )
        assert req.power_level_content_override.ban == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.invite == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.kick == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.redact == 50
        assert req.power_level_content_override.events_default == 0
        assert req.power_level_content_override.users_default == 0
        assert req.power_level_content_override.state_default == 50

    def test_arbitrary_event_type_override_does_not_skip_defaults(self) -> None:
        """
        Test that using unprotected event types in the power level content override does
        not affect said overrides
        """
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": {"events": {EventTypes.Message: 10}},
            "groups": [],
        }

        req = CreateManagedRoomRequest.model_validate(
            room_config, context={"room_creator": "@room_creator:example.com"}
        )
        assert len(req.power_level_content_override.users) == 0
        # By default, 3 event types are included in the events object, but we added one.
        # So there should be all 4 here
        self.assertDictEqual(
            req.power_level_content_override.events,
            {
                EventTypes.PowerLevels: CREATOR_POWER_LEVEL - 1,
                EventTypes.JoinRules: CREATOR_POWER_LEVEL - 1,
                EventTypes.GuestAccess: CREATOR_POWER_LEVEL - 1,
                EventTypes.Message: 10,
            },
        )
        assert req.power_level_content_override.ban == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.invite == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.kick == CREATOR_POWER_LEVEL - 1
        assert req.power_level_content_override.redact == 50
        assert req.power_level_content_override.events_default == 0
        assert req.power_level_content_override.users_default == 0
        assert req.power_level_content_override.state_default == 50

    def test_arbitrary_and_protected_event_type_override_can_error_appropriately(
        self,
    ) -> None:
        """
        Test that using both a protected and unprotected event type together in the
        power level content override does not allow the protected event type to be
        overridden
        """
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": {
                "events": {EventTypes.Message: 10, EventTypes.JoinRules: 100}
            },
            "groups": [],
        }

        # Since one of the event types is protected, this should error
        with pytest.raises(ValidationError):
            CreateManagedRoomRequest.model_validate(
                room_config, context={"room_creator": "@room_creator:example.com"}
            )

    def test_with_user_powerlevel_override(self) -> None:
        """
        Test that adding a user override power level operates as expected
        """
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": {"users": {"@alice:example.com": 100}},
            "groups": ["testgroup"],
        }

        req = CreateManagedRoomRequest.model_validate(
            room_config, context={"room_creator": "@room_creator:example.com"}
        )
        assert req.power_level_content_override.users_default == 0
        assert "@alice:example.com" in req.power_level_content_override.users
        assert req.power_level_content_override.users["@alice:example.com"] == 100
        # At the time this object is created, the room creator is not present in room
        # v10, this is added at the time of the request being processed. There should be
        # only the one we asked for
        assert len(req.power_level_content_override.users) == 1

    def test_not_adding_context_to_object_raises(self) -> None:
        """
        Test that not including the correct context for model validation causes failure.
        To be clear, this only will matter when the `users` object is present for the
        `power_level_content_override`. If the `users` object is not expressly being
        validated in the model, the context object is not required.
        """
        room_config = {
            "room_alias_name": "testroom",
            "name": "Test Room",
            "power_level_content_override": {"users": {"@alice:example.com": 100}},
            "groups": ["testgroup"],
        }

        # Note that the context is not added here. That data is needed to validate that
        # any users in the `user` override are not the room creator and should not be
        # skipped during power level validation
        with pytest.raises(ValidationError):
            CreateManagedRoomRequest.model_validate(room_config)

        # Also test that the wrong kind of object is not expected. It must be a dict
        # format
        with pytest.raises(ValidationError):
            CreateManagedRoomRequest.model_validate(
                room_config, context=["@room_creator:example.com"]
            )

        # Also test that "room_creator" being None is not allowed
        with pytest.raises(ValidationError):
            CreateManagedRoomRequest.model_validate(
                room_config, context={"room_creator": None}
            )
